import typing as t

import pytest

from sqlframe.base.catalog import CatalogMetadata, Column, Database, Function, Table
from sqlframe.redshift.session import RedshiftSession

pytest_plugins = ["tests.integration.fixtures"]
pytestmark = [
    pytest.mark.redshift,
    pytest.mark.xdist_group("redshift_tests"),
]


@pytest.fixture
def reset_database(redshift_session: RedshiftSession) -> t.Iterator[None]:
    yield
    redshift_session.catalog.setCurrentDatabase("public")


def test_current_catalog(redshift_session: RedshiftSession):
    assert redshift_session.catalog.currentCatalog() == "dev"


def test_set_current_catalog(redshift_session: RedshiftSession):
    with pytest.raises(NotImplementedError):
        redshift_session.catalog.setCurrentCatalog("catalog1")


def test_list_catalogs(redshift_session: RedshiftSession):
    assert sorted(redshift_session.catalog.listCatalogs(), key=lambda x: x.name) == [
        CatalogMetadata(name="dev", description=None)
    ]


def test_current_database(redshift_session: RedshiftSession):
    assert redshift_session.catalog.currentDatabase() == "public"


def test_set_current_database(redshift_session: RedshiftSession, reset_database):
    assert redshift_session.catalog.currentDatabase() == "public"
    redshift_session.catalog.setCurrentDatabase("db1")
    assert redshift_session.catalog.currentDatabase() == "db1"


def test_list_databases(redshift_session: RedshiftSession):
    assert sorted(redshift_session.catalog.listDatabases(), key=lambda x: (x.catalog, x.name)) == [
        Database(name="db1", catalog="dev", description=None, locationUri=""),
    ]


def test_list_databases_pattern(redshift_session: RedshiftSession):
    assert sorted(
        redshift_session.catalog.listDatabases("db*"), key=lambda x: (x.catalog, x.name)
    ) == [
        Database(name="db1", catalog="dev", description=None, locationUri=""),
    ]


def test_get_database_no_match(redshift_session: RedshiftSession):
    with pytest.raises(ValueError):
        assert redshift_session.catalog.getDatabase("nonexistent")


def test_get_database_name_only(redshift_session: RedshiftSession):
    assert redshift_session.catalog.getDatabase("db1") == Database(
        name="db1", catalog="dev", description=None, locationUri=""
    )


def test_get_database_name_and_catalog(redshift_session: RedshiftSession):
    assert redshift_session.catalog.getDatabase("catalog1.db1") == Database(
        name="db1", catalog="dev", description=None, locationUri=""
    )


def test_database_exists_does_exist(redshift_session: RedshiftSession):
    assert redshift_session.catalog.databaseExists("db1") is True


def test_database_exists_does_not_exist(redshift_session: RedshiftSession):
    assert redshift_session.catalog.databaseExists("nonexistent") is False


def test_list_tables_no_args(redshift_session: RedshiftSession):
    # no tables are returns since default schema is public
    assert (
        sorted(redshift_session.catalog.listTables(), key=lambda x: (x.catalog, x.database, x.name))
        == []
    )


def test_list_tables_db_no_catalog(redshift_session: RedshiftSession):
    assert sorted(
        redshift_session.catalog.listTables("db1"), key=lambda x: (x.catalog, x.database, x.name)
    ) == [
        Table(
            name="table1",
            catalog="dev",
            namespace=["db1"],
            description=None,
            tableType="MANAGED",
            isTemporary=False,
        )
    ]


def test_list_tables_db_and_catalog(redshift_session: RedshiftSession):
    assert sorted(
        redshift_session.catalog.listTables("dev.db1"),
        key=lambda x: (x.catalog, x.database, x.name),
    ) == [
        Table(
            name="table1",
            catalog="dev",
            namespace=["db1"],
            description=None,
            tableType="MANAGED",
            isTemporary=False,
        )
    ]


def test_list_tables_pattern(redshift_session: RedshiftSession):
    assert Table(
        name="table1",
        catalog="dev",
        namespace=["db1"],
        description=None,
        tableType="MANAGED",
        isTemporary=False,
    ) in redshift_session.catalog.listTables(pattern="tab*")


def test_get_table(redshift_session: RedshiftSession):
    assert redshift_session.catalog.getTable("dev.db1.table1") == Table(
        name="table1",
        catalog="dev",
        namespace=["db1"],
        description=None,
        tableType="MANAGED",
        isTemporary=False,
    )


def test_get_table_not_exists(redshift_session: RedshiftSession):
    with pytest.raises(ValueError):
        assert redshift_session.catalog.getTable("dev.db1.nonexistent")


def test_list_functions(redshift_session: RedshiftSession):
    # no functions are returned since default schema is public
    assert redshift_session.catalog.listFunctions() == []


def test_list_functions_pattern(redshift_session: RedshiftSession):
    assert redshift_session.catalog.listFunctions(dbName="db1", pattern="ad*") == [
        Function(
            name="add",
            catalog="dev",
            namespace=["db1"],
            description=None,
            className="",
            isTemporary=False,
        )
    ]


def test_function_exists_does_exist(redshift_session: RedshiftSession):
    assert redshift_session.catalog.functionExists("add", dbName="dev.db1") is True


def test_function_exists_does_not_exist(redshift_session: RedshiftSession):
    assert redshift_session.catalog.functionExists("nonexistent") is False


def test_get_function_exists(redshift_session: RedshiftSession):
    assert redshift_session.catalog.getFunction("dev.db1.add") == Function(
        name="add",
        catalog="dev",
        namespace=["db1"],
        description=None,
        className="",
        isTemporary=False,
    )


def test_get_function_not_exists(redshift_session: RedshiftSession):
    with pytest.raises(ValueError):
        assert redshift_session.catalog.getFunction("default.main.nonexistent")


def test_list_columns(redshift_session: RedshiftSession):
    assert sorted(redshift_session.catalog.listColumns("dev.db1.table1"), key=lambda x: x.name) == [
        Column(
            name="id",
            description=None,
            dataType="INT",
            nullable=True,
            isPartition=False,
            isBucket=False,
        ),
        Column(
            name="name",
            description=None,
            dataType="STRING",
            nullable=True,
            isPartition=False,
            isBucket=False,
        ),
    ]


def test_list_columns_use_db_name(redshift_session: RedshiftSession):
    assert sorted(
        redshift_session.catalog.listColumns("table1", dbName="dev.db1"), key=lambda x: x.name
    ) == [
        Column(
            name="id",
            description=None,
            dataType="INT",
            nullable=True,
            isPartition=False,
            isBucket=False,
        ),
        Column(
            name="name",
            description=None,
            dataType="STRING",
            nullable=True,
            isPartition=False,
            isBucket=False,
        ),
    ]


def test_table_exists_table_name_only(redshift_session: RedshiftSession):
    assert redshift_session.catalog.tableExists("dev.db1.table1") is True


def test_table_exists_table_name_and_db_name(redshift_session: RedshiftSession):
    assert redshift_session.catalog.tableExists("table1", dbName="dev.db1") is True


def test_table_not_exists(redshift_session: RedshiftSession):
    assert redshift_session.catalog.tableExists("nonexistent") is False


def test_create_external_table(redshift_session: RedshiftSession):
    with pytest.raises(NotImplementedError):
        redshift_session.catalog.createExternalTable("table1", "tests.public", "path/to/table")


def test_create_table(redshift_session: RedshiftSession):
    with pytest.raises(NotImplementedError):
        redshift_session.catalog.createTable("table1", "tests.public")


def test_drop_temp_view(redshift_session: RedshiftSession):
    with pytest.raises(NotImplementedError):
        redshift_session.catalog.dropTempView("view1")


def test_drop_global_temp_view(redshift_session: RedshiftSession):
    with pytest.raises(NotImplementedError):
        redshift_session.catalog.dropGlobalTempView("view1")


def test_register_function(redshift_session: RedshiftSession):
    with pytest.raises(NotImplementedError):
        redshift_session.catalog.registerFunction("function1", lambda x: x)


def test_is_cached(redshift_session: RedshiftSession):
    with pytest.raises(NotImplementedError):
        redshift_session.catalog.isCached("table1")


def test_cache_table(redshift_session: RedshiftSession):
    with pytest.raises(NotImplementedError):
        redshift_session.catalog.cacheTable("table1")


def test_uncache_table(redshift_session: RedshiftSession):
    with pytest.raises(NotImplementedError):
        redshift_session.catalog.uncacheTable("table1")


def test_clear_cache(redshift_session: RedshiftSession):
    with pytest.raises(NotImplementedError):
        redshift_session.catalog.clearCache()


def test_refresh_table(redshift_session: RedshiftSession):
    with pytest.raises(NotImplementedError):
        redshift_session.catalog.refreshTable("table1")


def test_recover_partitions(redshift_session: RedshiftSession):
    with pytest.raises(NotImplementedError):
        redshift_session.catalog.recoverPartitions("table1")


def test_refresh_by_path(redshift_session: RedshiftSession):
    with pytest.raises(NotImplementedError):
        redshift_session.catalog.refreshByPath("path/to/table")



from __future__ import annotations

import typing as t

import pytest

from sqlframe.base import types
from sqlframe.base.exceptions import DataFrameDiffError, SchemaDiffError
from sqlframe.testing import assertDataFrameEqual, assertSchemaEqual

if t.TYPE_CHECKING:
    from sqlframe.base.dataframe import _BaseSession


pytest_plugins = ["tests.integration.fixtures"]


def test_df_equal_no_diff(get_session: t.Callable[[], _BaseSession]):
    session = get_session()
    df1 = session.createDataFrame(data=[("1", 1000), ("2", 3000)], schema=["id", "amount"])
    df2 = session.createDataFrame(data=[("1", 1000), ("2", 3000)], schema=["id", "amount"])
    assertDataFrameEqual(df1, df2)
    df1 = session.createDataFrame(data=[("1", 0.1), ("2", 3.23)], schema=["id", "amount"])
    df2 = session.createDataFrame(data=[("1", 0.109), ("2", 3.23)], schema=["id", "amount"])
    assertDataFrameEqual(df1, df2, rtol=1e-1)  # pass, DataFrames are approx equal by rtol
    df1 = session.createDataFrame(data=[(1, 1000), (2, 3000)], schema=["id", "amount"])
    list_of_rows = [types.Row(1, 1000), types.Row(2, 3000)]  # type: ignore
    assertDataFrameEqual(df1, list_of_rows)  # pass, actual and expected data are equal


def test_df_equal_diff(get_session: t.Callable[[], _BaseSession]):
    session = get_session()
    df1 = session.createDataFrame(data=[("1", 1000), ("2", 3000)], schema=["id", "amount"])
    df2 = session.createDataFrame(data=[("1", 1000), ("2", 3000)], schema=["id", "amount"])
    df2 = df2.withColumn("amount", df2["amount"] + 1)
    with pytest.raises(DataFrameDiffError):
        assertDataFrameEqual(df1, df2)
    df1 = session.createDataFrame(
        data=[("1", 1000.00), ("2", 3000.00), ("3", 2000.00)], schema=["id", "amount"]
    )
    df2 = session.createDataFrame(
        data=[("1", 1001.00), ("2", 3000.00), ("3", 2003.00)], schema=["id", "amount"]
    )
    with pytest.raises(DataFrameDiffError):
        assertDataFrameEqual(df1, df2)


def test_schema_equal_no_diff(get_session: t.Callable[[], _BaseSession]):
    session = get_session()
    schema1 = session.createDataFrame(
        data=[("1", 1000), ("2", 3000)], schema=["id", "amount"]
    ).schema
    schema2 = session.createDataFrame(
        data=[("1", 1000), ("2", 3000)], schema=["id", "amount"]
    ).schema
    assertSchemaEqual(schema1, schema2)
    s1 = types.StructType(
        [types.StructField("names", types.ArrayType(types.DoubleType(), True), True)]
    )
    s2 = types.StructType(
        [types.StructField("names", types.ArrayType(types.DoubleType(), True), True)]
    )
    assertSchemaEqual(s1, s2)  # pass, schemas are identical


def test_schema_equal_diff(get_session: t.Callable[[], _BaseSession]):
    session = get_session()
    schema1 = session.createDataFrame(
        data=[("1", 1000), ("2", 3000)], schema=["id", "amount"]
    ).schema
    schema2 = session.createDataFrame(
        data=[("1", 1000), ("2", 3000)], schema=["diff", "amount"]
    ).schema
    with pytest.raises(SchemaDiffError):
        assertSchemaEqual(schema1, schema2)
    df1 = session.createDataFrame(data=[(1, 1000), (2, 3000)], schema=["id", "number"])
    df2 = session.createDataFrame(data=[("1", 1000), ("2", 5000)], schema=["id", "amount"])
    with pytest.raises(SchemaDiffError):
        assertSchemaEqual(df1.schema, df2.schema)



from __future__ import annotations

import typing as t

from sqlframe.base.udf import _BaseUDFRegistration

if t.TYPE_CHECKING:
    from sqlframe.base._typing import DataTypeOrString, UserDefinedFunctionLike
    from sqlframe.duckdb.session import DuckDBSession


class DuckDBUDFRegistration(_BaseUDFRegistration["DuckDBSession"]):
    def register(  # type: ignore
        self,
        name: str,
        f: t.Union[t.Callable[..., t.Any], UserDefinedFunctionLike],
        returnType: t.Optional[DataTypeOrString] = None,
    ) -> UserDefinedFunctionLike:
        self.sparkSession._conn.create_function(name, f, return_type=returnType)  # type: ignore



import typing as t

import pytest
from pyspark.errors import AnalysisException

from sqlframe.spark.catalog import CatalogMetadata, Column, Database, Function, Table
from sqlframe.spark.session import SparkSession

pytest_plugins = ["tests.integration.fixtures"]


@pytest.fixture
def reset_database(spark_session: SparkSession) -> t.Iterator[None]:
    yield
    spark_session.catalog.setCurrentDatabase("db1")


def test_current_catalog(spark_session: SparkSession):
    assert spark_session.catalog.currentCatalog() == "spark_catalog"


def test_list_catalogs(spark_session: SparkSession):
    assert sorted(spark_session.catalog.listCatalogs(), key=lambda x: x.name) == [
        CatalogMetadata(name="spark_catalog", description=None)
    ]


def test_current_database(spark_session: SparkSession):
    assert spark_session.catalog.currentDatabase() == "db1"


def test_set_current_database(spark_session: SparkSession, reset_database):
    assert spark_session.catalog.currentDatabase() == "db1"
    spark_session.catalog.setCurrentDatabase("default")
    assert spark_session.catalog.currentDatabase() == "default"


def test_list_databases(spark_session: SparkSession):
    assert sorted(
        [
            Database(name=x.name, catalog=x.catalog, description="", locationUri="")
            for x in spark_session.catalog.listDatabases()
        ],
        key=lambda x: (x.catalog, x.name),
    ) == [
        Database(name="db1", catalog="spark_catalog", description="", locationUri=""),
        Database(name="default", catalog="spark_catalog", description="", locationUri=""),
    ]


def test_list_databases_pattern(spark_session: SparkSession):
    assert sorted(
        [
            Database(name=x.name, catalog=x.catalog, description="", locationUri="")
            for x in spark_session.catalog.listDatabases("db*")
        ],
        key=lambda x: (x.catalog, x.name),
    ) == [
        Database(name="db1", catalog="spark_catalog", description="", locationUri=""),
    ]


def test_get_database_no_match(spark_session: SparkSession):
    with pytest.raises(AnalysisException):
        assert spark_session.catalog.getDatabase("nonexistent")


def test_get_database_name_only(spark_session: SparkSession):
    database = spark_session.catalog.getDatabase("db1")
    assert database.name == "db1"
    assert database.catalog == "spark_catalog"


def test_database_exists_does_exist(spark_session: SparkSession):
    assert spark_session.catalog.databaseExists("db1") is True


def test_database_exists_does_not_exist(spark_session: SparkSession):
    assert spark_session.catalog.databaseExists("nonexistent") is False


def test_list_tables_no_args(spark_session: SparkSession):
    assert sorted(
        # Filter out temporary tables from TPCDS test
        [x for x in spark_session.catalog.listTables() if x.tableType != "TEMPORARY"],
        key=lambda x: (x.catalog, x.database, x.name),
    ) == [
        Table(
            name="table1",
            catalog="spark_catalog",
            namespace=["db1"],
            description=None,
            tableType="MANAGED",
            isTemporary=False,
        )
    ]


def test_list_tables_db_no_catalog(spark_session: SparkSession):
    assert sorted(
        # Filter out temporary tables from TPCDS test
        [x for x in spark_session.catalog.listTables("db1") if x.tableType != "TEMPORARY"],
        key=lambda x: (x.catalog or "", x.database, x.name),
    ) == [
        Table(
            name="table1",
            catalog="spark_catalog",
            namespace=["db1"],
            description=None,
            tableType="MANAGED",
            isTemporary=False,
        )
    ]


def test_list_tables_db_and_catalog(spark_session: SparkSession):
    assert sorted(
        # Filter out temporary tables from TPCDS test
        [
            x
            for x in spark_session.catalog.listTables("spark_catalog.db1")
            if x.tableType != "TEMPORARY"
        ],
        key=lambda x: (x.catalog or "", x.database, x.name),
    ) == [
        Table(
            name="table1",
            catalog="spark_catalog",
            namespace=["db1"],
            description=None,
            tableType="MANAGED",
            isTemporary=False,
        )
    ]


def test_list_tables_pattern(spark_session: SparkSession):
    assert sorted(
        spark_session.catalog.listTables(pattern="tab*"),
        key=lambda x: (x.catalog, x.database, x.name),
    ) == [
        Table(
            name="table1",
            catalog="spark_catalog",
            namespace=["db1"],
            description=None,
            tableType="MANAGED",
            isTemporary=False,
        )
    ]


def test_get_table(spark_session: SparkSession):
    assert spark_session.catalog.getTable("spark_catalog.db1.table1") == Table(
        name="table1",
        catalog="spark_catalog",
        namespace=["db1"],
        description=None,
        tableType="MANAGED",
        isTemporary=False,
    )


def test_get_table_not_exists(spark_session: SparkSession):
    with pytest.raises(AnalysisException):
        assert spark_session.catalog.getTable("spark_catalog.db1.nonexistent")


def test_list_functions(spark_session: SparkSession):
    assert "add" in [x.name for x in spark_session.catalog.listFunctions()]


def test_list_functions_pattern(spark_session: SparkSession):
    assert "add" in [x.name for x in spark_session.catalog.listFunctions(pattern="ad*")]


def test_function_exists_does_exist(spark_session: SparkSession):
    assert spark_session.catalog.functionExists("add") is True


def test_function_exists_does_not_exist(spark_session: SparkSession):
    assert spark_session.catalog.functionExists("nonexistent") is False


def test_get_function_exists(spark_session: SparkSession):
    function = spark_session.catalog.getFunction("add")
    assert function.name == "add"
    assert function.catalog is None
    assert function.namespace is None
    assert function.isTemporary is True


def test_get_function_not_exists(spark_session: SparkSession):
    with pytest.raises(AnalysisException):
        assert spark_session.catalog.getFunction("default.main.nonexistent")


def test_list_columns(spark_session: SparkSession):
    assert sorted(
        spark_session.catalog.listColumns("spark_catalog.db1.table1"), key=lambda x: x.name
    ) == [
        Column(
            name="id",
            description=None,
            dataType="INT",
            nullable=True,
            isPartition=False,
            isBucket=False,
        ),
        Column(
            name="name",
            description=None,
            dataType="STRING",
            nullable=True,
            isPartition=False,
            isBucket=False,
        ),
    ]


def test_list_columns_use_db_name(spark_session: SparkSession):
    assert sorted(
        spark_session.catalog.listColumns("table1", dbName="db1"), key=lambda x: x.name
    ) == [
        Column(
            name="id",
            description=None,
            dataType="INT",
            nullable=True,
            isPartition=False,
            isBucket=False,
        ),
        Column(
            name="name",
            description=None,
            dataType="STRING",
            nullable=True,
            isPartition=False,
            isBucket=False,
        ),
    ]


def test_table_exists_table_name_only(spark_session: SparkSession):
    assert spark_session.catalog.tableExists("spark_catalog.db1.table1") is True


def test_table_exists_table_name_and_db_name(spark_session: SparkSession):
    assert spark_session.catalog.tableExists("table1", dbName="db1") is True


def test_table_not_exists(spark_session: SparkSession):
    assert spark_session.catalog.tableExists("nonexistent") is False



# This code is based on code from Apache Spark under the license found in the LICENSE file located in the 'sqlframe' folder.
from __future__ import annotations

import typing as t

if t.TYPE_CHECKING:
    from sqlframe.base._typing import DataTypeOrString, UserDefinedFunctionLike
    from sqlframe.base.session import _BaseSession

    SESSION = t.TypeVar("SESSION", bound=_BaseSession)
else:
    SESSION = t.TypeVar("SESSION")


class _BaseUDFRegistration(t.Generic[SESSION]):
    def __init__(self, sparkSession: SESSION):
        self.sparkSession = sparkSession

    def register(
        self,
        name: str,
        f: t.Union[t.Callable[..., t.Any], UserDefinedFunctionLike],
        returnType: t.Optional[DataTypeOrString] = None,
    ) -> UserDefinedFunctionLike:
        raise NotImplementedError

    def registerJavaFunction(
        self,
        name: str,
        javaClassName: str,
        returnType: t.Optional[DataTypeOrString] = None,
    ) -> None:
        raise NotImplementedError

    def registerJavaUDAF(self, name: str, javaClassName: str) -> None:
        raise NotImplementedError



import pytest

from sqlframe.snowflake import Column as SnowflakeColumn
from sqlframe.snowflake import Row as SnowflakeRow
from sqlframe.snowflake import (
    SnowflakeCatalog,
    SnowflakeDataFrame,
    SnowflakeDataFrameNaFunctions,
    SnowflakeDataFrameReader,
    SnowflakeDataFrameStatFunctions,
    SnowflakeDataFrameWriter,
    SnowflakeGroupedData,
    SnowflakeSession,
    SnowflakeUDFRegistration,
)
from sqlframe.snowflake import Window as SnowflakeWindow
from sqlframe.snowflake import WindowSpec as SnowflakeWindowSpec
from sqlframe.snowflake import functions as SnowflakeF
from sqlframe.snowflake import types as SnowflakeTypes


@pytest.mark.forked
def test_activate_snowflake(check_pyspark_imports):
    check_pyspark_imports(
        "snowflake",
        sqlf_session=SnowflakeSession,
        sqlf_catalog=SnowflakeCatalog,
        sqlf_column=SnowflakeColumn,
        sqlf_dataframe=SnowflakeDataFrame,
        sqlf_grouped_data=SnowflakeGroupedData,
        sqlf_window=SnowflakeWindow,
        sqlf_window_spec=SnowflakeWindowSpec,
        sqlf_functions=SnowflakeF,
        sqlf_types=SnowflakeTypes,
        sqlf_udf_registration=SnowflakeUDFRegistration,
        sqlf_dataframe_reader=SnowflakeDataFrameReader,
        sqlf_dataframe_writer=SnowflakeDataFrameWriter,
        sqlf_dataframe_na_functions=SnowflakeDataFrameNaFunctions,
        sqlf_dataframe_stat_functions=SnowflakeDataFrameStatFunctions,
        sqlf_row=SnowflakeRow,
    )



# This code is based on code from Apache Spark under the license found in the LICENSE file located in the 'sqlframe' folder.

from __future__ import annotations

import sys
import typing as t

import sqlglot as sg
from sqlglot import exp

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from sqlframe.base.mixins.readwriter_mixins import PandasLoaderMixin, PandasWriterMixin
from sqlframe.base.readerwriter import (
    _BaseDataFrameReader,
    _BaseDataFrameWriter,
)
from sqlframe.base.util import normalize_string

if t.TYPE_CHECKING:
    from sqlframe.databricks.session import DatabricksSession  # noqa
    from sqlframe.databricks.dataframe import DatabricksDataFrame  # noqa


class DatabricksDataFrameReader(
    PandasLoaderMixin["DatabricksSession", "DatabricksDataFrame"],
    _BaseDataFrameReader["DatabricksSession", "DatabricksDataFrame"],
):
    pass


class DatabricksDataFrameWriter(
    PandasWriterMixin["DatabricksSession", "DatabricksDataFrame"],
    _BaseDataFrameWriter["DatabricksSession", "DatabricksDataFrame"],
):
    def saveAsTable(
        self,
        name: str,
        format: t.Optional[str] = None,
        mode: t.Optional[str] = None,
        partitionBy: t.Optional[t.Union[str, t.List[str]]] = None,
        clusterBy: t.Optional[t.Union[str, t.List[str]]] = None,
        **options,
    ) -> Self:
        if format is not None:
            raise NotImplementedError("Providing Format in the save as table is not supported")
        exists, replace, mode = None, None, mode or str(self._mode)
        if mode == "append":
            return self.insertInto(name)
        if mode == "ignore":
            exists = True
        if mode == "overwrite":
            replace = True
        name = normalize_string(name, from_dialect="input", is_table=True)

        properties: t.List[exp.Expression] = []
        if partitionBy is not None:
            if isinstance(partitionBy, str):
                partition_by = [partitionBy]
            else:
                partition_by = partitionBy
            properties.append(
                exp.PartitionedByProperty(
                    this=exp.Tuple(expressions=list(map(sg.to_identifier, partition_by)))
                )
            )
        if clusterBy is not None:
            if isinstance(clusterBy, str):
                cluster_by = [clusterBy]
            else:
                cluster_by = clusterBy
            properties.append(
                exp.Cluster(
                    expressions=[exp.Tuple(expressions=list(map(sg.to_identifier, cluster_by)))]
                )
            )

        properties.extend(
            exp.Property(this=sg.to_identifier(name), value=exp.convert(value))
            for name, value in (options or {}).items()
        )

        output_expression_container = exp.Create(
            this=exp.to_table(name, dialect=self._session.input_dialect),
            kind="TABLE",
            exists=exists,
            replace=replace,
            properties=exp.Properties(expressions=properties),
        )
        df = self._df.copy(output_expression_container=output_expression_container)
        if self._session._has_connection:
            df.collect()
        return self.copy(_df=df)



import pytest

from sqlframe import activate

pytest_plugins = ["tests.common_fixtures"]


@pytest.mark.forked
def test_activate_with_connection(function_scoped_postgres):
    cursor = function_scoped_postgres.cursor()
    cursor.execute('CREATE SCHEMA "activate_test"')
    cursor.execute('CREATE TABLE "activate_test"."test" (a INT)')
    cursor.execute('INSERT INTO "activate_test"."test" VALUES (1)')
    activate("postgres", conn=function_scoped_postgres)
    from pyspark.sql import SparkSession

    spark = SparkSession.builder.appName("test").getOrCreate()

    df = spark.table("activate_test.test").select("`a`")
    assert df.collect() == [(1,)]


@pytest.mark.forked
def test_activate_with_connection_and_input_dialect(function_scoped_postgres):
    cursor = function_scoped_postgres.cursor()
    cursor.execute('CREATE SCHEMA "activate_test"')
    cursor.execute('CREATE TABLE "activate_test"."test" (a INT)')
    cursor.execute('INSERT INTO "activate_test"."test" VALUES (1)')
    activate(
        "postgres", conn=function_scoped_postgres, config={"sqlframe.input.dialect": "postgres"}
    )
    from pyspark.sql import SparkSession

    spark = SparkSession.builder.appName("test").getOrCreate()

    df = spark.table("activate_test.test").select('"a"')
    assert df.collect() == [(1,)]



import pytest

from sqlframe import activate
from sqlframe.databricks import Column as DatabricksColumn
from sqlframe.databricks import (
    DatabricksCatalog,
    DatabricksDataFrame,
    DatabricksDataFrameNaFunctions,
    DatabricksDataFrameReader,
    DatabricksDataFrameStatFunctions,
    DatabricksDataFrameWriter,
    DatabricksGroupedData,
    DatabricksSession,
    DatabricksUDFRegistration,
)
from sqlframe.databricks import Row as DatabricksRow
from sqlframe.databricks import Window as DatabricksWindow
from sqlframe.databricks import WindowSpec as DatabricksWindowSpec
from sqlframe.databricks import functions as DatabricksF
from sqlframe.databricks import types as DatabricksTypes


@pytest.mark.forked
def test_activate_databricks(check_pyspark_imports):
    check_pyspark_imports(
        "databricks",
        sqlf_session=DatabricksSession,
        sqlf_catalog=DatabricksCatalog,
        sqlf_column=DatabricksColumn,
        sqlf_dataframe=DatabricksDataFrame,
        sqlf_grouped_data=DatabricksGroupedData,
        sqlf_window=DatabricksWindow,
        sqlf_window_spec=DatabricksWindowSpec,
        sqlf_functions=DatabricksF,
        sqlf_types=DatabricksTypes,
        sqlf_udf_registration=DatabricksUDFRegistration,
        sqlf_dataframe_reader=DatabricksDataFrameReader,
        sqlf_dataframe_writer=DatabricksDataFrameWriter,
        sqlf_dataframe_na_functions=DatabricksDataFrameNaFunctions,
        sqlf_dataframe_stat_functions=DatabricksDataFrameStatFunctions,
        sqlf_row=DatabricksRow,
    )


@pytest.mark.forked
def test_activate_databricks_default_dataset():
    activate("databricks")
    from pyspark.sql import SparkSession

    assert SparkSession == DatabricksSession



from __future__ import annotations

import typing as t

from sqlframe.base.udf import _BaseUDFRegistration

if t.TYPE_CHECKING:
    from sqlframe.bigquery.session import BigQuerySession


class BigQueryUDFRegistration(_BaseUDFRegistration["BigQuerySession"]): ...



from __future__ import annotations

import typing as t

import pytest

from sqlframe.base import types
from sqlframe.base.types import Row
from sqlframe.bigquery import BigQuerySession
from sqlframe.postgres import PostgresSession

if t.TYPE_CHECKING:
    from sqlframe.base.session import _BaseSession

pytest_plugins = ["tests.integration.fixtures"]


def test_columnt_eq_null_safe(get_session: t.Callable[[], _BaseSession], get_func):
    session = get_session()
    df1 = session.createDataFrame([Row(id=1, value="foo"), Row(id=2, value=None)])
    lit = get_func("lit", session)
    values = df1.select(
        df1["value"] == "foo",
        df1["value"].eqNullSafe("foo"),
        df1["value"].eqNullSafe(None),
        df1["id"].eqNullSafe(1),
        df1["id"] == lit(None),
        df1["id"].eqNullSafe(None),
    ).collect()

    assert values[0] == (True, True, False, True, None, False)
    assert values[1] == (None, False, True, False, None, False)


def test_column_get_item_array(get_session: t.Callable[[], _BaseSession], get_func):
    session = get_session()
    lit = get_func("lit", session)
    assert session.range(1).select(lit(["a", "b", "c"]).getItem(0).alias("value")).first()[0] == "a"


def test_column_get_item_map(get_session: t.Callable[[], _BaseSession], get_func):
    session = get_session()
    lit = get_func("lit", session)
    if not isinstance(session, (PostgresSession, BigQuerySession)):
        assert session.range(1).select(lit({"key": "value"}).getItem("key")).first()[0] == "value"



import pytest

from sqlframe import activate


@pytest.mark.forked
def test_activate_with_connection():
    import duckdb

    connector = duckdb.connect()
    connector.execute('CREATE SCHEMA "memory"."activate_test"')
    connector.execute('CREATE TABLE "memory"."activate_test"."test" (a INT)')
    connector.execute('INSERT INTO "memory"."activate_test"."test" VALUES (1)')
    activate("duckdb", conn=connector)
    from pyspark.sql import SparkSession

    spark = SparkSession.builder.appName("test").getOrCreate()

    df = spark.table("memory.activate_test.test").select("`a`")
    assert df.collect() == [(1,)]


@pytest.mark.forked
def test_activate_with_connection_and_input_dialect():
    import duckdb

    connector = duckdb.connect()
    connector.execute('CREATE SCHEMA "memory"."activate_test"')
    connector.execute('CREATE TABLE "memory"."activate_test"."test" (a INT)')
    connector.execute('INSERT INTO "memory"."activate_test"."test" VALUES (1)')
    activate("duckdb", conn=connector, config={"sqlframe.input.dialect": "duckdb"})
    from pyspark.sql import SparkSession

    spark = SparkSession.builder.appName("test").getOrCreate()

    df = spark.table("memory.activate_test.test").select('"a"')
    assert df.collect() == [(1,)]



import typing as t

import pytest
from sqlglot import expressions as exp

from sqlframe.base.exceptions import UnsupportedOperationError
from sqlframe.standalone import functions as F
from sqlframe.standalone.dataframe import StandaloneDataFrame

pytest_plugins = ["tests.common_fixtures", "tests.unit.standalone.fixtures"]


def test_hash_select_expression(standalone_employee: StandaloneDataFrame):
    expression = exp.select("cola").from_("table")
    assert standalone_employee._create_hash_from_expression(expression) == "t17051938"


def test_columns(standalone_employee: StandaloneDataFrame):
    assert standalone_employee.columns == ["employee_id", "fname", "lname", "age", "store_id"]


def test_cache(standalone_employee: StandaloneDataFrame, compare_sql: t.Callable):
    df = standalone_employee.select("fname").cache()
    expected_statements = [
        "DROP VIEW IF EXISTS `t31563989`",
        "CACHE LAZY TABLE `t31563989` OPTIONS('storageLevel' = 'MEMORY_AND_DISK') AS SELECT CAST(`a1`.`fname` AS STRING) AS `fname` FROM VALUES (1, 'Jack', 'Shephard', 37, 1), (2, 'John', 'Locke', 65, 1), (3, 'Kate', 'Austen', 37, 2), (4, 'Claire', 'Littleton', 27, 2), (5, 'Hugo', 'Reyes', 29, 100) AS `a1`(`employee_id`, `fname`, `lname`, `age`, `store_id`)",
        "SELECT `t31563989`.`fname` AS `fname` FROM `t31563989` AS `t31563989`",
    ]
    compare_sql(df, expected_statements)


def test_persist_default(standalone_employee: StandaloneDataFrame, compare_sql: t.Callable):
    df = standalone_employee.select("fname").persist()
    expected_statements = [
        "DROP VIEW IF EXISTS `t31563989`",
        "CACHE LAZY TABLE `t31563989` OPTIONS('storageLevel' = 'MEMORY_AND_DISK_SER') AS SELECT CAST(`a1`.`fname` AS STRING) AS `fname` FROM VALUES (1, 'Jack', 'Shephard', 37, 1), (2, 'John', 'Locke', 65, 1), (3, 'Kate', 'Austen', 37, 2), (4, 'Claire', 'Littleton', 27, 2), (5, 'Hugo', 'Reyes', 29, 100) AS `a1`(`employee_id`, `fname`, `lname`, `age`, `store_id`)",
        "SELECT `t31563989`.`fname` AS `fname` FROM `t31563989` AS `t31563989`",
    ]
    compare_sql(df, expected_statements)


def test_persist_storagelevel(standalone_employee: StandaloneDataFrame, compare_sql: t.Callable):
    df = standalone_employee.select("fname").persist("DISK_ONLY_2")
    expected_statements = [
        "DROP VIEW IF EXISTS `t31563989`",
        "CACHE LAZY TABLE `t31563989` OPTIONS('storageLevel' = 'DISK_ONLY_2') AS SELECT CAST(`a1`.`fname` AS STRING) AS `fname` FROM VALUES (1, 'Jack', 'Shephard', 37, 1), (2, 'John', 'Locke', 65, 1), (3, 'Kate', 'Austen', 37, 2), (4, 'Claire', 'Littleton', 27, 2), (5, 'Hugo', 'Reyes', 29, 100) AS `a1`(`employee_id`, `fname`, `lname`, `age`, `store_id`)",
        "SELECT `t31563989`.`fname` AS `fname` FROM `t31563989` AS `t31563989`",
    ]
    compare_sql(df, expected_statements)


def test_with_column_duplicate_alias(standalone_employee: StandaloneDataFrame):
    df = standalone_employee.withColumn("fName", F.col("age").cast("string"))
    assert df.columns == ["employee_id", "fname", "lname", "age", "store_id"]
    # Make sure that the new columns is added with an alias to `fname`
    assert (
        df.sql(pretty=False)
        == "SELECT `a1`.`employee_id` AS `employee_id`, CAST(`a1`.`age` AS STRING) AS `fname`, CAST(`a1`.`lname` AS STRING) AS `lname`, `a1`.`age` AS `age`, `a1`.`store_id` AS `store_id` FROM VALUES (1, 'Jack', 'Shephard', 37, 1), (2, 'John', 'Locke', 65, 1), (3, 'Kate', 'Austen', 37, 2), (4, 'Claire', 'Littleton', 27, 2), (5, 'Hugo', 'Reyes', 29, 100) AS `a1`(`employee_id`, `fname`, `lname`, `age`, `store_id`)"
    )


def test_with_columns(standalone_employee: StandaloneDataFrame):
    df = standalone_employee.withColumns({"new_col1": F.col("age"), "new_col2": F.col("store_id")})
    assert df.columns == [
        "employee_id",
        "fname",
        "lname",
        "age",
        "store_id",
        "new_col1",
        "new_col2",
    ]
    assert (
        df.sql(pretty=False)
        == "SELECT `a1`.`employee_id` AS `employee_id`, CAST(`a1`.`fname` AS STRING) AS `fname`, CAST(`a1`.`lname` AS STRING) AS `lname`, `a1`.`age` AS `age`, `a1`.`store_id` AS `store_id`, `a1`.`age` AS `new_col1`, `a1`.`store_id` AS `new_col2` FROM VALUES (1, 'Jack', 'Shephard', 37, 1), (2, 'John', 'Locke', 65, 1), (3, 'Kate', 'Austen', 37, 2), (4, 'Claire', 'Littleton', 27, 2), (5, 'Hugo', 'Reyes', 29, 100) AS `a1`(`employee_id`, `fname`, `lname`, `age`, `store_id`)"
    )


def test_transform(standalone_employee: StandaloneDataFrame):
    def cast_all_to_int(input_df):
        return input_df.select([F.col(col_name).cast("int") for col_name in input_df.columns])

    def sort_columns_asc(input_df):
        return input_df.select(*sorted(input_df.columns))

    df = standalone_employee.transform(cast_all_to_int).transform(sort_columns_asc)
    assert df.columns == ["age", "employee_id", "fname", "lname", "store_id"]
    assert df.sql(pretty=False, optimize=False).endswith(  # type: ignore
        "SELECT CAST(`employee_id` AS INT) AS `employee_id`, CAST(`fname` AS INT) AS `fname`, CAST(`lname` AS INT) AS `lname`, CAST(`age` AS INT) AS `age`, CAST(`store_id` AS INT) AS `store_id` FROM `t51718876`) SELECT `age`, `employee_id`, `fname`, `lname`, `store_id` FROM `t16881256`"
    )


# https://github.com/eakmanrq/sqlframe/issues/19
def test_with_column_dual_expression(standalone_employee: StandaloneDataFrame):
    df1 = standalone_employee.withColumn("new_col1", standalone_employee.age)
    df2 = df1.withColumn("new_col2", standalone_employee.store_id)
    assert df2.columns == [
        "employee_id",
        "fname",
        "lname",
        "age",
        "store_id",
        "new_col1",
        "new_col2",
    ]
    assert (
        df2.sql(pretty=False)
        == "SELECT `a1`.`employee_id` AS `employee_id`, CAST(`a1`.`fname` AS STRING) AS `fname`, CAST(`a1`.`lname` AS STRING) AS `lname`, `a1`.`age` AS `age`, `a1`.`store_id` AS `store_id`, `a1`.`age` AS `new_col1`, `a1`.`store_id` AS `new_col2` FROM VALUES (1, 'Jack', 'Shephard', 37, 1), (2, 'John', 'Locke', 65, 1), (3, 'Kate', 'Austen', 37, 2), (4, 'Claire', 'Littleton', 27, 2), (5, 'Hugo', 'Reyes', 29, 100) AS `a1`(`employee_id`, `fname`, `lname`, `age`, `store_id`)"
    )


def test_where_expr(standalone_employee: StandaloneDataFrame):
    df = standalone_employee.where("fname = 'Jack' AND age = 37")
    assert df.columns == ["employee_id", "fname", "lname", "age", "store_id"]
    assert (
        df.sql(pretty=False)
        == "SELECT `a1`.`employee_id` AS `employee_id`, CAST(`a1`.`fname` AS STRING) AS `fname`, CAST(`a1`.`lname` AS STRING) AS `lname`, `a1`.`age` AS `age`, `a1`.`store_id` AS `store_id` FROM VALUES (1, 'Jack', 'Shephard', 37, 1), (2, 'John', 'Locke', 65, 1), (3, 'Kate', 'Austen', 37, 2), (4, 'Claire', 'Littleton', 27, 2), (5, 'Hugo', 'Reyes', 29, 100) AS `a1`(`employee_id`, `fname`, `lname`, `age`, `store_id`) WHERE `a1`.`age` = 37 AND CAST(`a1`.`fname` AS STRING) = 'Jack'"
    )


def test_missing_method(standalone_employee: StandaloneDataFrame):
    with pytest.raises(
        UnsupportedOperationError, match="Tried to call a column which is unexpected.*"
    ):
        standalone_employee.missing_method("blah")


def test_expand_star(standalone_employee: StandaloneDataFrame):
    assert (
        standalone_employee.select("*").sql(pretty=False, optimize=False)
        == "WITH `t51718876` AS (SELECT CAST(`employee_id` AS INT) AS `employee_id`, CAST(`fname` AS STRING) AS `fname`, CAST(`lname` AS STRING) AS `lname`, CAST(`age` AS INT) AS `age`, CAST(`store_id` AS INT) AS `store_id` FROM VALUES (1, 'Jack', 'Shephard', 37, 1), (2, 'John', 'Locke', 65, 1), (3, 'Kate', 'Austen', 37, 2), (4, 'Claire', 'Littleton', 27, 2), (5, 'Hugo', 'Reyes', 29, 100) AS `a1`(`employee_id`, `fname`, `lname`, `age`, `store_id`)) SELECT `employee_id`, `fname`, `lname`, `age`, `store_id` FROM `t51718876`"
    )


def test_expand_star_table_alias(standalone_employee: StandaloneDataFrame):
    assert (
        standalone_employee.alias("blah").select("blah.*").sql(pretty=False, optimize=False)
        == "WITH `t51718876` AS (SELECT CAST(`employee_id` AS INT) AS `employee_id`, CAST(`fname` AS STRING) AS `fname`, CAST(`lname` AS STRING) AS `lname`, CAST(`age` AS INT) AS `age`, CAST(`store_id` AS INT) AS `store_id` FROM VALUES (1, 'Jack', 'Shephard', 37, 1), (2, 'John', 'Locke', 65, 1), (3, 'Kate', 'Austen', 37, 2), (4, 'Claire', 'Littleton', 27, 2), (5, 'Hugo', 'Reyes', 29, 100) AS `a1`(`employee_id`, `fname`, `lname`, `age`, `store_id`)), `t37842204` AS (SELECT `employee_id`, `fname`, `lname`, `age`, `store_id` FROM `t51718876`) SELECT `t37842204`.`employee_id`, `t37842204`.`fname`, `t37842204`.`lname`, `t37842204`.`age`, `t37842204`.`store_id` FROM `t37842204`"
    )


def test_lineage(standalone_employee: StandaloneDataFrame):
    assert (
        standalone_employee.lineage("age").source.sql()
        == "SELECT a1.age AS age FROM (VALUES (1, 'Jack', 'Shephard', 37, 1), (2, 'John', 'Locke', 65, 1), (3, 'Kate', 'Austen', 37, 2), (4, 'Claire', 'Littleton', 27, 2), (5, 'Hugo', 'Reyes', 29, 100)) AS a1(employee_id, fname, lname, age, store_id)"
    )
    assert (
        standalone_employee.session.sql("SELECT * FROM employee").lineage("age").source.sql()
        == "SELECT employee.age AS age FROM employee AS employee"
    )



from __future__ import annotations

import typing as t

from sqlglot import exp

from sqlframe.base.session import _BaseSession
from sqlframe.postgres.catalog import PostgresCatalog
from sqlframe.postgres.dataframe import PostgresDataFrame
from sqlframe.postgres.readwriter import (
    PostgresDataFrameReader,
    PostgresDataFrameWriter,
)
from sqlframe.postgres.udf import PostgresUDFRegistration

if t.TYPE_CHECKING:
    from psycopg2.extensions import connection as psycopg2_connection

    from sqlframe.base.types import Row
else:
    psycopg2_connection = t.Any


class PostgresSession(
    _BaseSession[  # type: ignore
        PostgresCatalog,
        PostgresDataFrameReader,
        PostgresDataFrameWriter,
        PostgresDataFrame,
        psycopg2_connection,
        PostgresUDFRegistration,
    ],
):
    _catalog = PostgresCatalog
    _reader = PostgresDataFrameReader
    _writer = PostgresDataFrameWriter
    _df = PostgresDataFrame
    _udf_registration = PostgresUDFRegistration

    def __init__(self, conn: t.Optional[psycopg2_connection] = None):
        if not hasattr(self, "_conn"):
            super().__init__(conn)
            self._execute("CREATE EXTENSION IF NOT EXISTS fuzzystrmatch")
            self._execute("""CREATE OR REPLACE FUNCTION pg_temp.try_to_timestamp(input_text TEXT, format TEXT)
RETURNS TIMESTAMP AS $$
BEGIN
    RETURN TO_TIMESTAMP(input_text, format);
EXCEPTION WHEN OTHERS THEN
    RETURN NULL;
END;
$$ LANGUAGE plpgsql;""")

    def _collect(
        self,
        expressions: t.Union[str, exp.Expression, t.List[str], t.List[exp.Expression]],
        *,
        quote_identifiers: bool = True,
        skip_normalization: bool = False,
        skip_rows: bool = False,
    ) -> t.List[Row]:
        from psycopg2 import ProgrammingError

        try:
            return super()._collect(
                expressions,
                quote_identifiers=quote_identifiers,
                skip_normalization=skip_normalization,
            )
        except ProgrammingError as e:
            if "no results to fetch" in str(e):
                return []
            raise e

    class Builder(_BaseSession.Builder):
        DEFAULT_EXECUTION_DIALECT = "postgres"

        @property
        def session(self) -> PostgresSession:
            return PostgresSession(**self._session_kwargs)

        def getOrCreate(self) -> PostgresSession:
            return super().getOrCreate()  # type: ignore

    builder = Builder()



# This code is based on code from Apache Spark under the license found in the LICENSE file located in the 'sqlframe' folder.

from __future__ import annotations

import fnmatch
import json
import typing as t

from sqlglot import exp, parse_one

from sqlframe.base.catalog import Column, Function, _BaseCatalog
from sqlframe.base.mixins.catalog_mixins import (
    GetCurrentCatalogFromFunctionMixin,
    GetCurrentDatabaseFromFunctionMixin,
    ListCatalogsFromInfoSchemaMixin,
    ListDatabasesFromInfoSchemaMixin,
    ListTablesFromInfoSchemaMixin,
    SetCurrentCatalogFromUseMixin,
    SetCurrentDatabaseFromUseMixin,
)
from sqlframe.base.util import normalize_string, schema_, to_schema

if t.TYPE_CHECKING:
    from sqlframe.databricks.session import DatabricksSession  # noqa
    from sqlframe.databricks.dataframe import DatabricksDataFrame  # noqa


class DatabricksCatalog(
    GetCurrentCatalogFromFunctionMixin["DatabricksSession", "DatabricksDataFrame"],
    GetCurrentDatabaseFromFunctionMixin["DatabricksSession", "DatabricksDataFrame"],
    ListDatabasesFromInfoSchemaMixin["DatabricksSession", "DatabricksDataFrame"],
    ListCatalogsFromInfoSchemaMixin["DatabricksSession", "DatabricksDataFrame"],
    SetCurrentDatabaseFromUseMixin["DatabricksSession", "DatabricksDataFrame"],
    ListTablesFromInfoSchemaMixin["DatabricksSession", "DatabricksDataFrame"],
    _BaseCatalog["DatabricksSession", "DatabricksDataFrame"],
):
    CURRENT_CATALOG_EXPRESSION: exp.Expression = exp.func("current_catalog")
    UPPERCASE_INFO_SCHEMA = True

    def setCurrentCatalog(self, catalogName: str) -> None:
        self.session._collect(
            exp.Use(
                kind=exp.Var(this=exp.to_identifier("CATALOG")),
                this=exp.parse_identifier(catalogName, dialect=self.session.input_dialect),
            ),
            quote_identifiers=False,
        )

    def listFunctions(
        self, dbName: t.Optional[str] = None, pattern: t.Optional[str] = None
    ) -> t.List[Function]:
        """
        Returns a t.List of functions registered in the specified database.

        .. versionadded:: 3.4.0

        Parameters
        ----------
        dbName : str
            name of the database to t.List the functions.
            ``dbName`` can be qualified with catalog name.
        pattern : str
            The pattern that the function name needs to match.

            .. versionchanged: 3.5.0
                Adds ``pattern`` argument.

        Returns
        -------
        t.List
            A t.List of :class:`Function`.

        Notes
        -----
        If no database is specified, the current database and catalog
        are used. This API includes all temporary functions.

        Examples
        --------
        >>> spark.catalog.t.listFunctions()
        [Function(name=...

        >>> spark.catalog.t.listFunctions(pattern="to_*")
        [Function(name=...

        >>> spark.catalog.t.listFunctions(pattern="*not_existing_func*")
        []
        """
        if dbName is None:
            schema = schema_(
                db=exp.parse_identifier(
                    self.currentDatabase(), dialect=self.session.output_dialect
                ),
                catalog=exp.parse_identifier(
                    self.currentCatalog(), dialect=self.session.output_dialect
                ),
            )
        else:
            dbName = normalize_string(dbName, from_dialect="input", is_schema=True)
            schema = to_schema(dbName, dialect=self.session.input_dialect)
            if not schema.catalog:
                schema.set(
                    "catalog",
                    exp.parse_identifier(
                        self.currentCatalog(), dialect=self.session.output_dialect
                    ),
                )
        current_catalog = self.currentCatalog()
        self.session._collect(f"USE CATALOG {schema.catalog}")

        query = parse_one(
            f"""SHOW USER FUNCTIONS IN {schema.sql(dialect=self.session.input_dialect)}""",
            dialect=self.session.input_dialect,
        )
        functions = [
            Function(
                name=normalize_string(
                    x["function"].split(".")[-1], from_dialect="execution", to_dialect="output"
                ),
                catalog=normalize_string(
                    schema.catalog, from_dialect="execution", to_dialect="output"
                ),
                namespace=[
                    normalize_string(schema.db, from_dialect="execution", to_dialect="output")
                ],
                description=None,
                className="",
                isTemporary=False,
            )
            for x in self.session._collect(query)
        ]
        if pattern:
            normalized_pattern = normalize_string(
                pattern, from_dialect="input", to_dialect="output", is_pattern=True
            )
            functions = [x for x in functions if fnmatch.fnmatch(x.name, normalized_pattern)]
        self.session._collect(f"USE CATALOG {current_catalog}")
        return functions

    def get_columns(self, table: exp.Table | str) -> t.Dict[str, exp.DataType]:
        table = (
            normalize_string(table, from_dialect="input", is_table=True)
            if isinstance(table, str)
            else table
        )
        table = exp.to_table(table, dialect=self.session.input_dialect)
        if not table.catalog:
            table.set(
                "catalog",
                exp.parse_identifier(
                    normalize_string(
                        self.currentCatalog(), from_dialect="output", to_dialect="input"
                    ),
                    dialect=self.session.input_dialect,
                ),
            )
        if not table.db:
            table.set(
                "db",
                exp.parse_identifier(
                    normalize_string(
                        self.currentDatabase(), from_dialect="output", to_dialect="input"
                    ),
                    dialect=self.session.input_dialect,
                ),
            )
        sql = f"DESCRIBE TABLE {table.sql(dialect=self.session.input_dialect)}"
        results = self.session._collect(sql)
        return {
            normalize_string(
                row["col_name"],
                from_dialect="execution",
                to_dialect="output",
                is_column=True,
            ): exp.DataType.build(
                normalize_string(
                    row["data_type"],
                    from_dialect="execution",
                    to_dialect="output",
                    is_datatype=True,
                ),
                dialect=self.session.output_dialect,
                udt=True,
            )
            for row in results
            if row["data_type"] != "" and row["data_type"] != "data_type"
        }

    def listColumns(
        self, tableName: str, dbName: t.Optional[str] = None, include_temp: bool = False
    ) -> t.List[Column]:
        """Returns a t.List of columns for the given table/view in the specified database.

        .. versionadded:: 2.0.0

        Parameters
        ----------
        tableName : str
            name of the table to t.List columns.

            .. versionchanged:: 3.4.0
               Allow ``tableName`` to be qualified with catalog name when ``dbName`` is None.

        dbName : str, t.Optional
            name of the database to find the table to t.List columns.

        Returns
        -------
        t.List
            A t.List of :class:`Column`.

        Notes
        -----
        The order of arguments here is different from that of its JVM counterpart
        because Python does not support method overloading.

        If no database is specified, the current database and catalog
        are used. This API includes all temporary views.

        Examples
        --------
        >>> _ = spark.sql("DROP TABLE IF EXISTS tbl1")
        >>> _ = spark.sql("CREATE TABLE tblA (name STRING, age INT) USING parquet")
        >>> spark.catalog.t.listColumns("tblA")
        [Column(name='name', description=None, dataType='string', nullable=True, ...
        >>> _ = spark.sql("DROP TABLE tblA")
        """
        tableName = normalize_string(
            tableName, from_dialect=self.session.input_dialect, is_table=True
        )
        dbName = (
            normalize_string(dbName, from_dialect=self.session.input_dialect, is_schema=True)
            if dbName
            else None
        )
        if df := self.session.temp_views.get(tableName):
            return [
                Column(
                    name=x,
                    description=None,
                    dataType="",
                    nullable=True,
                    isPartition=False,
                    isBucket=False,
                )
                for x in df.columns
            ]

        table = exp.to_table(tableName, dialect=self.session.input_dialect)
        schema = to_schema(dbName, dialect=self.session.input_dialect) if dbName else None
        if not table.db:
            if schema and schema.db:
                table.set("db", schema.args["db"])
            else:
                current_database = normalize_string(
                    self.currentDatabase(),
                    from_dialect=self.session.output_dialect,
                    to_dialect=self.session.input_dialect,
                )
                table.set(
                    "db",
                    exp.parse_identifier(current_database, dialect=self.session.input_dialect),
                )
        if not table.catalog:
            if schema and schema.catalog:
                table.set("catalog", schema.args["catalog"])
            else:
                current_catalog = normalize_string(
                    self.currentCatalog(),
                    from_dialect=self.session.output_dialect,
                    to_dialect=self.session.input_dialect,
                )
                table.set(
                    "catalog",
                    exp.parse_identifier(current_catalog, dialect=self.session.input_dialect),
                )
        sql = f"DESCRIBE TABLE {'.'.join(part.name for part in table.parts)}"
        results = self.session._collect(sql)

        is_partition = False
        partitions = set([])
        for row in results:
            if row["col_name"] == "# Partition Information":
                is_partition = True
            if is_partition and row["data_type"] != "" and row["data_type"] != "data_type":
                partitions.add(row["col_name"])

        columns = []
        for row in results:
            if row["data_type"] == "" or row["data_type"] == "data_type":
                break
            columns.append(
                Column(
                    name=normalize_string(
                        row["col_name"],
                        from_dialect=self.session.execution_dialect,
                        to_dialect=self.session.output_dialect,
                    ),
                    description=row["comment"],
                    dataType=normalize_string(
                        row["data_type"],
                        from_dialect=self.session.execution_dialect,
                        to_dialect=self.session.output_dialect,
                        is_datatype=True,
                    ),
                    nullable=True,
                    isPartition=True if row["col_name"] in partitions else False,
                    isBucket=False,
                )
            )

        return columns



import datetime

import pytest

from sqlframe.base import types
from sqlframe.postgres import PostgresDataFrame, PostgresSession

pytest_plugins = ["tests.integration.fixtures"]


@pytest.fixture()
def postgres_datatypes(postgres_session: PostgresSession) -> PostgresDataFrame:
    return postgres_session.createDataFrame(
        [
            (
                1,
                2.0,
                "foo",
                [1, 2, 3],
                datetime.date(2022, 1, 1),
                datetime.datetime(2022, 1, 1, 0, 0, 0),
                datetime.datetime(2022, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc),
                True,
            )
        ],
        [
            "bigint_col",
            "double_col",
            "string_col",
            "array<bigint>_col",
            "date_col",
            "timestamp_col",
            "timestamptz_col",
            "boolean_col",
        ],
    )


def test_print_schema_basic(postgres_employee: PostgresDataFrame, capsys):
    postgres_employee.printSchema()
    captured = capsys.readouterr()
    assert (
        captured.out.strip()
        == """
root
 |-- employee_id: int (nullable = true)
 |-- fname: string (nullable = true)
 |-- lname: string (nullable = true)
 |-- age: int (nullable = true)
 |-- store_id: int (nullable = true)""".strip()
    )


def test_print_schema_nested(postgres_datatypes: PostgresDataFrame, capsys):
    postgres_datatypes.printSchema()
    captured = capsys.readouterr()
    # array does not include type
    assert (
        captured.out.strip()
        == """
root
 |-- bigint_col: bigint (nullable = true)
 |-- double_col: double (nullable = true)
 |-- string_col: string (nullable = true)
 |-- `array<bigint>_col`: array<bigint> (nullable = true)
 |    |-- element: bigint (nullable = true)
 |-- date_col: date (nullable = true)
 |-- timestamp_col: timestamp (nullable = true)
 |-- timestamptz_col: timestamp (nullable = true)
 |-- boolean_col: boolean (nullable = true)""".strip()
    )


def test_schema(postgres_employee: PostgresDataFrame):
    assert postgres_employee.schema == types.StructType(
        [
            types.StructField(
                "employee_id",
                types.IntegerType(),
            ),
            types.StructField(
                "fname",
                types.StringType(),
            ),
            types.StructField(
                "lname",
                types.StringType(),
            ),
            types.StructField(
                "age",
                types.IntegerType(),
            ),
            types.StructField(
                "store_id",
                types.IntegerType(),
            ),
        ]
    )


def test_schema_nested(postgres_datatypes: PostgresDataFrame):
    assert isinstance(postgres_datatypes.schema, types.StructType)
    struct_fields = list(postgres_datatypes.schema)
    assert len(struct_fields) == 8
    assert struct_fields[0].name == "bigint_col"
    assert struct_fields[0].dataType == types.LongType()
    assert struct_fields[1].name == "double_col"
    assert struct_fields[1].dataType == types.DoubleType()
    assert struct_fields[2].name == "string_col"
    assert struct_fields[2].dataType == types.StringType()
    assert struct_fields[3].name == "`array<bigint>_col`"
    assert struct_fields[3].dataType == types.ArrayType(
        types.LongType(),
    )
    assert struct_fields[4].name == "date_col"
    assert struct_fields[4].dataType == types.DateType()
    assert struct_fields[5].name == "timestamp_col"
    assert struct_fields[5].dataType == types.TimestampType()
    assert struct_fields[6].name == "timestamptz_col"
    assert struct_fields[6].dataType == types.TimestampType()
    assert struct_fields[7].name == "boolean_col"
    assert struct_fields[7].dataType == types.BooleanType()



from __future__ import annotations

import pathlib
import typing as t

import pytest

from sqlframe.base.session import _BaseSession
from sqlframe.base.types import Row
from sqlframe.duckdb.session import DuckDBSession

if t.TYPE_CHECKING:
    from sqlframe.base.dataframe import _BaseDataFrame

pytest_plugins = ["tests.integration.fixtures"]


@pytest.fixture
def cleanup_employee_df(
    get_engine_df: t.Callable[[str], _BaseDataFrame],
) -> t.Iterator[_BaseDataFrame]:
    df = get_engine_df("employee")
    df.session._execute("DROP TABLE IF EXISTS insert_into_employee")
    df.session._execute("DROP TABLE IF EXISTS save_as_table_employee")
    yield df
    df.session._execute("DROP TABLE IF EXISTS insert_into_employee")
    df.session._execute("DROP TABLE IF EXISTS save_as_table_employee")


def test_write_json(get_engine_df: t.Callable[[str], _BaseDataFrame], tmp_path: pathlib.Path):
    df_employee = get_engine_df("employee")
    temp_json = str(tmp_path / "employee.json")
    df_employee.write.json(temp_json)
    df2 = df_employee.session.read.json(temp_json)
    assert df2.collect() == df_employee.collect()


def test_write_json_append(get_session: t.Callable[[], _BaseSession], tmp_path: pathlib.Path):
    session = get_session()
    temp_json = tmp_path / "test.json"
    df1 = session.createDataFrame([(1,)])
    df2 = session.createDataFrame([(2,)])
    if isinstance(session, DuckDBSession):
        with pytest.raises(NotImplementedError):
            df1.write.json(str(temp_json), mode="append")
    else:
        df1.write.json(str(temp_json), mode="append")
        df2.write.json(str(temp_json), mode="append")
        df_result = session.read.json(str(temp_json))
        assert df_result.collect() == [Row(_1=1), Row(_1=2)]


def test_write_json_ignore(
    get_engine_df: t.Callable[[str], _BaseDataFrame], tmp_path: pathlib.Path
):
    df_employee = get_engine_df("employee")
    temp_json = tmp_path / "employee.json"
    temp_json.touch()
    df_employee.write.json(str(temp_json), mode="ignore")
    df2 = df_employee.session.read.json(str(temp_json))
    assert df2.collect() == []


def test_write_json_error(
    get_engine_df: t.Callable[[str], _BaseDataFrame], tmp_path: pathlib.Path, caplog
):
    df_employee = get_engine_df("employee")
    temp_json = tmp_path / "employee.json"
    temp_json.touch()
    with pytest.raises(FileExistsError):
        df_employee.write.json(temp_json, mode="error")


def test_write_parquet(get_engine_df: t.Callable[[str], _BaseDataFrame], tmp_path: pathlib.Path):
    df_employee = get_engine_df("employee")
    temp_parquet = str(tmp_path / "employee.parquet")
    df_employee.write.parquet(temp_parquet)
    df2 = df_employee.session.read.parquet(temp_parquet)
    assert df2.collect() == df_employee.collect()


def test_write_parquet_ignore(
    get_engine_df: t.Callable[[str], _BaseDataFrame], tmp_path: pathlib.Path
):
    df_employee = get_engine_df("employee")
    temp_parquet = str(tmp_path / "employee.parquet")
    df_empty = df_employee.session.createDataFrame([])
    # Write nothing
    df_empty.write.parquet(temp_parquet)
    # Try writing actual data
    df_employee.write.parquet(temp_parquet, mode="ignore")
    df2 = df_employee.session.read.parquet(temp_parquet)
    # Check that second write did not happen
    assert df2.collect() == []


def test_write_parquet_error(
    get_engine_df: t.Callable[[str], _BaseDataFrame], tmp_path: pathlib.Path, caplog
):
    df_employee = get_engine_df("employee")
    temp_parquet = tmp_path / "employee.parquet"
    temp_parquet.touch()
    with pytest.raises(FileExistsError):
        df_employee.write.json(temp_parquet, mode="error")


def test_write_parquet_unsupported_modes(
    get_engine_df: t.Callable[[str], _BaseDataFrame], tmp_path: pathlib.Path
):
    df_employee = get_engine_df("employee")
    temp_json = tmp_path / "employee.parquet"
    with pytest.raises(NotImplementedError):
        df_employee.write.parquet(str(temp_json), mode="append")


def test_write_csv(get_engine_df: t.Callable[[str], _BaseDataFrame], tmp_path: pathlib.Path):
    df_employee = get_engine_df("employee")
    temp_csv = str(tmp_path / "employee.csv")
    df_employee.write.csv(temp_csv)
    df2 = df_employee.session.read.csv(temp_csv)
    assert df2.collect() == df_employee.collect()


def test_write_csv_append(get_session: t.Callable[[], _BaseSession], tmp_path: pathlib.Path):
    session = get_session()
    temp_csv = tmp_path / "test.csv"
    df1 = session.createDataFrame([(1,)])
    df2 = session.createDataFrame([(2,)])
    if isinstance(session, DuckDBSession):
        with pytest.raises(NotImplementedError):
            df1.write.csv(str(temp_csv), mode="append")
    else:
        df1.write.csv(str(temp_csv), mode="append")
        df2.write.csv(str(temp_csv), mode="append")
        df_result = session.read.csv(str(temp_csv))
        assert df_result.collect() == [Row(_1=1), Row(_1=2)]


def test_write_csv_ignore(get_engine_df: t.Callable[[str], _BaseDataFrame], tmp_path: pathlib.Path):
    df_employee = get_engine_df("employee")
    temp_csv = str(tmp_path / "employee.csv")
    df1 = df_employee.session.createDataFrame([(1,)])
    df2 = df_employee.session.createDataFrame([(2,)])
    # Write nothing
    df1.write.csv(temp_csv)
    # Try writing actual data
    df2.write.csv(temp_csv, mode="ignore")
    df_result = df_employee.session.read.csv(temp_csv)
    # Check that second write did not happen
    assert df_result.collect() == df1.collect()


def test_write_csv_error(get_engine_df: t.Callable[[str], _BaseDataFrame], tmp_path: pathlib.Path):
    df_employee = get_engine_df("employee")
    temp_csv = tmp_path / "employee.csv"
    temp_csv.touch()
    with pytest.raises(FileExistsError):
        df_employee.write.json(temp_csv, mode="error")


def test_save_as_table(cleanup_employee_df: _BaseDataFrame, caplog):
    df_employee = cleanup_employee_df
    df_employee.write.saveAsTable("save_as_table_employee")
    df2 = df_employee.session.read.table("save_as_table_employee")
    assert sorted(df2.collect()) == sorted(df_employee.collect())


def test_insertInto(cleanup_employee_df: _BaseDataFrame, caplog):
    df_employee = cleanup_employee_df
    df = df_employee.session.createDataFrame(
        [(9, "Sayid", "Jarrah", 40, 1)], ["id", "first_name", "last_name", "age", "store_id"]
    )
    df_employee.write.saveAsTable("insert_into_employee")
    df.write.insertInto("insert_into_employee")
    df2 = df_employee.session.read.table("insert_into_employee")
    assert sorted(df_employee.union(df).collect()) == sorted(df2.collect())



from __future__ import annotations

import importlib
import sys
import typing as t
from unittest.mock import MagicMock

if t.TYPE_CHECKING:
    from sqlframe.base.session import CONN

ENGINE_TO_PREFIX = {
    "bigquery": "BigQuery",
    "duckdb": "DuckDB",
    "postgres": "Postgres",
    "redshift": "Redshift",
    "snowflake": "Snowflake",
    "spark": "Spark",
    "databricks": "Databricks",
    "standalone": "Standalone",
}

NAME_TO_FILE_OVERRIDE = {
    "DataFrameNaFunctions": "dataframe",
    "DataFrameStatFunctions": "dataframe",
    "DataFrameReader": "readwriter",
    "DataFrameWriter": "readwriter",
    "GroupedData": "group",
    "SparkSession": "session",
    "WindowSpec": "window",
    "UDFRegistration": "udf",
}

ACTIVATE_CONFIG = {}


def activate(
    engine: t.Optional[str] = None,
    conn: t.Optional[CONN] = None,
    config: t.Optional[t.Dict[str, t.Any]] = None,
) -> None:
    import sqlframe
    from sqlframe import testing

    pyspark_mock = MagicMock()
    pyspark_mock.__file__ = "pyspark"
    sys.modules["pyspark"] = pyspark_mock
    pyspark_mock.testing = testing
    sys.modules["pyspark.testing"] = testing
    if conn:
        ACTIVATE_CONFIG["sqlframe.conn"] = conn
    for key, value in (config or {}).items():
        ACTIVATE_CONFIG[key] = value
    if not engine:
        return
    engine = engine.lower()
    if engine not in ENGINE_TO_PREFIX:
        raise ValueError(
            f"Unsupported engine {engine}. Supported engines are {', '.join(ENGINE_TO_PREFIX)}"
        )
    prefix = ENGINE_TO_PREFIX[engine]
    engine_module = importlib.import_module(f"sqlframe.{engine}")

    sys.modules["pyspark.sql"] = engine_module
    pyspark_mock.sql = engine_module
    types = engine_module.__dict__.copy()
    resolved_files = set()
    for name, obj in types.items():
        if name.startswith(prefix) or name in [
            "Column",
            "Window",
            "WindowSpec",
            "functions",
            "types",
        ]:
            name_without_prefix = name.replace(prefix, "")
            if name_without_prefix == "Session":
                name_without_prefix = "SparkSession"
            setattr(engine_module, name_without_prefix, obj)
            file = NAME_TO_FILE_OVERRIDE.get(name_without_prefix, name_without_prefix).lower()
            engine_file = importlib.import_module(f"sqlframe.{engine}.{file}")
            if engine_file not in resolved_files:
                sys.modules[f"pyspark.sql.{file}"] = engine_file
                resolved_files.add(engine_file)
            setattr(engine_file, name_without_prefix, obj)


def deactivate() -> None:
    pyspark_imports = [k for k in sys.modules if k.startswith("pyspark")]

    for k, v in sys.modules.copy().items():
        if k in pyspark_imports:
            del sys.modules[k]
    # Try importing the pyspark imports again and see if pyspark is installed and therefore available
    # if not then nothing will change
    for k in pyspark_imports:
        try:
            sys.modules[k] = importlib.import_module(k)
        except ImportError:
            pass
    ACTIVATE_CONFIG.clear()
